
//cpu controller state machine
//Eric R. Williamson
//JUL2005

`timescale 1ns / 1ps

module controller_fsm	(	clock,
				tb_read_rom,
				tb_read_rand,
				reset,
				read_rom,
				read_rom_done,
				Op,
				pc_reset,
				PCWrite,
				PCInc,
				PCWriteCond,
				IorD,
				MemRead,
				MemWrite,
				IRWrite,
				MemtoReg,
				PCSource,
				TargetWrite,
				ALUOp,
				ALUSelB,
				ALUSelA,
				RegWrite,
				RegDst,
			);

input tb_read_rom, tb_read_rand, reset, read_rom_done;
output read_rom;
reg read_rom;

input clock;
input [5:0] Op;
output pc_reset, PCWrite, PCInc, PCWriteCond, IorD, MemRead;
output MemWrite, IRWrite, MemtoReg;
output [1:0] PCSource;
output TargetWrite;
output [1:0] ALUOp;
output [1:0] ALUSelB;
output ALUSelA, RegWrite, RegDst;

reg PCWrite, PCInc, PCWriteCond, IorD, MemRead;
reg MemWrite, IRWrite, MemtoReg;
reg [1:0] PCSource;
reg TargetWrite;
reg [1:0] ALUOp;
reg [1:0] ALUSelB;
reg ALUSelA, RegWrite, RegDst;
reg pc_reset;

//Op code parameter assignments
parameter 
	RType = 6'b000000,
	jmp = 6'b000010,
	beq = 6'b100100,
	lw = 6'b100011,
	sw = 6'b101011;


//State Assignments
parameter 
	RESET = 10,
	READ_ROM = 11,
	READ_RAND = 12,
	INST_FETCH = 0,
	INST_DECODE_REG_FETCH = 1,
	MEM_ADDR_COMP = 2,
	MEM_READ = 3,
	WRITE_BCK_STEP = 4,
	MEM_WRITE = 5,
	EXECUTION = 6,
	RTYPE_COMPLETE = 7,
	BRANCH_COMPLETE = 8,
	JUMP_COMPLETE = 9;

reg [3:0] present_state;
reg [3:0] next_state;


covergroup cg @(present_state);
FSM: coverpoint present_state {
	bins ps_b1[] = {[0:12]};
}

//  option.cross_auto_bin_max = 20;
//cross MST, FSM1, FSM2, FSM3;

endgroup

cg cg_inst = new();


//---------
//FSM
//---------

always @(clock)
begin: PS_LOGIC
	case(present_state)
		RESET: begin
			pc_reset = 1'b1;
			next_state = RESET;
			if(tb_read_rand == 1'b1) next_state = READ_RAND;
			if(tb_read_rom == 1'b1) next_state = READ_ROM;
		end
		READ_ROM: begin
			pc_reset = 1'b1;
			MemWrite = 1'b1;
			read_rom = 1'b1;
			if(read_rom_done == 1'b1) begin
				read_rom = 1'b0;
				next_state = INST_FETCH;
			end
			else
				next_state = READ_ROM;
		end
		READ_RAND: begin
			pc_reset = 1'b1;
			MemWrite = 1'b1;
			read_rom = 1'b0;
			if(tb_read_rand !== 1'b1) begin
				next_state = INST_FETCH;
			end
			else
				next_state = READ_RAND;
		end
		INST_FETCH: begin
			pc_reset = 1'b0;
			next_state = INST_DECODE_REG_FETCH;
			PCWrite = 1'b0;
			PCInc = 1'b0;
			PCWriteCond = 1'b0;
			IorD = 1'b0;
			MemRead = 1'b1;
			MemWrite = 1'b0;
			IRWrite = 1'b1;
			MemtoReg = 1'b0;
			PCSource = 2'b00;
			TargetWrite = 1'b0;
			ALUOp = 2'b00;
			ALUSelB = 2'b01;
			ALUSelA = 1'b0;
			RegWrite = 1'b0;
			RegDst = 1'b0;
		end
		INST_DECODE_REG_FETCH: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b0;
			PCWriteCond = 1'b0;
			IorD = 1'b0;
			MemRead = 1'b0;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b00;
			TargetWrite = 1'b1;
			ALUOp = 2'b00;
			ALUSelB = 2'b11;
			ALUSelA = 1'b0;
			RegWrite = 1'b0;
			RegDst = 1'b0;
			if(Op == lw || Op == sw)
				next_state = MEM_ADDR_COMP;
			if(Op == RType)
				next_state = EXECUTION;
			if(Op == beq)
				next_state = BRANCH_COMPLETE;
			if(Op == jmp)
				next_state = JUMP_COMPLETE;
			if(tb_read_rand == 1'b1)
				next_state = READ_RAND;
			if(tb_read_rom == 1'b1)
				next_state = READ_ROM;
		end
		MEM_ADDR_COMP: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b0;
			PCWriteCond = 1'b0;
			IorD = 1'b0;
			MemRead = 1'b0;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b00;
			TargetWrite = 1'b0;
			ALUOp = 2'b00;
			ALUSelB = 2'b10;
			ALUSelA = 1'b1;
			RegWrite = 1'b0;
			RegDst = 1'b0;
			if(Op == lw)
				next_state = MEM_READ;
			if(Op == sw)
				next_state = MEM_WRITE;
		end
		MEM_READ: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b0;
			PCWriteCond = 1'b0;
			IorD = 1'b1;
			MemRead = 1'b1;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b00;
			TargetWrite = 1'b0;
			ALUOp = 2'b00;
			ALUSelB = 2'b10;
			ALUSelA = 1'b1;
			RegWrite = 1'b0;
			RegDst = 1'b0;
			next_state = WRITE_BCK_STEP;
		end
		WRITE_BCK_STEP: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b1;
			PCWriteCond = 1'b0;
			IorD = 1'b1;
			MemRead = 1'b1;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b1;
			PCSource = 2'b00;
			TargetWrite = 1'b0;
			ALUOp = 2'b00;
			ALUSelB = 2'b10;
			ALUSelA = 1'b1;
			RegWrite = 1'b1;
			RegDst = 1'b0;
			next_state = INST_FETCH;
		end
		MEM_WRITE: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b1;
			PCWriteCond = 1'b0;
			IorD = 1'b1;
			MemRead = 1'b0;
			MemWrite = 1'b1;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b00;
			TargetWrite = 1'b0;
			ALUOp = 2'b00;
			ALUSelB = 2'b10;
			ALUSelA = 1'b1;
			RegWrite = 1'b0;
			RegDst = 1'b0;
			next_state = INST_FETCH;
		end
		EXECUTION: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b0;
			PCWriteCond = 1'b0;
			IorD = 1'b0;
			MemRead = 1'b0;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b00;
			TargetWrite = 1'b0;
			ALUOp = 2'b10;
			ALUSelB = 2'b00;
			ALUSelA = 1'b1;
			RegWrite = 1'b0;
			RegDst = 1'b1;
			next_state = RTYPE_COMPLETE;
		end
		RTYPE_COMPLETE: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b1;
			PCWriteCond = 1'b0;
			IorD = 1'b0;
			MemRead = 1'b0;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b00;
			TargetWrite = 1'b0;
			ALUOp = 2'b10;
			ALUSelB = 2'b00;
			ALUSelA = 1'b1;
			RegWrite = 1'b1;
			RegDst = 1'b1;
			next_state = INST_FETCH;
		end
		BRANCH_COMPLETE: begin
			pc_reset = 1'b0;
			PCWrite = 1'b0;
			PCInc = 1'b1;
			PCWriteCond = 1'b1;
			IorD = 1'b0;
			MemRead = 1'b0;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b01;
			TargetWrite = 1'b0;
			ALUOp = 2'b01;
			ALUSelB = 2'b00;
			ALUSelA = 1'b1;
			RegWrite = 1'b0;
			RegDst = 1'b0;
			next_state = INST_FETCH;
		end
		JUMP_COMPLETE: begin
			pc_reset = 1'b0;
			PCWrite = 1'b1;
			PCInc = 1'b0;
			PCWriteCond = 1'b0;
			IorD = 1'b0;
			MemRead = 1'b0;
			MemWrite = 1'b0;
			IRWrite = 1'b0;
			MemtoReg = 1'b0;
			PCSource = 2'b10;
			TargetWrite = 1'b0;
			ALUOp = 2'b00;
			ALUSelB = 2'b00;
			ALUSelA = 1'b0;
			RegWrite = 1'b0;
			RegDst = 1'b0;
			next_state = INST_FETCH;
		end

		default: next_state = present_state;
	endcase
end


always @(posedge clock)
begin: NS_LOGIC
	if(reset == 1'b1)
		present_state = RESET;
	else
		present_state = next_state;
end

endmodule

